define(function (require) {
  'use strict';

  // Imports
  var U = require('utils'),
      Tool = require('tool'),
      Graph = require('graph'),
      Line = require('line'),
      Point = require('point'),
      Diamond = require('diamond'),
      Polygon = require('polygon'),
      Text = require('text'),
      HSlider = require('hslider'),
      Checkbox = require('checkbox'),
      Button = require('button'),
      _ = require('underscore');

   // UI
  var tool,
      graph,
      xReadoutText, yReadoutText,
      constraintsLabel, objFunctionLabel,
      constraintsText, objFunctionText,
      aSlider, bSlider, cSlider,
      resetButton, removeCb;

  var A = 0, OPER = 1, B = 2, SIGN = 3, C = 4;  // text indices

  var constraints = [],
      objFunction,
      ipLine,  // intersect polygon line/points
      ipPoints,
      polygon,

      dragEnabled = true,
      polygonIntersected = false,
      stage = 0,
      ci = 2;

  var MAX_CONSTRAINTS = 8,
      BIG_NUM = 100000,
      INIT = -10;  // initialize graph shapes outside of clipping bounds

  // State variables (JSInput)
  var state = {
    // TODO: Add state bindings
  };

  //// OBJECT CLASSES

  // point modification to track x and y values when setting XY position
  Point.prototype.setPos = function(x, y) {
    this.setXY(x, y);
    this.x = x;
    this.y = y;
  };

  // Constraint and Obj Function Lines

  function LPLine(cssPrefix) {  // prototype
    this.a = 0.0;
    this.b = 0.0;
    this.c = 0.0;
    this.xs = this.ys = this.xe = this.ye = INIT;
    this.line = new Line(graph, INIT, INIT, INIT, INIT, cssPrefix + '-line');
  }

  LPLine.prototype.setLine = function() {
    if ((this.a === 0.0 && this.b === 0.0)) {
      this.xs = this.ys = this.xe = this.ye = INIT;
    }
    else if (this.a === 0.0) {  // horizontal
      this.xs = graph.xMin;
      this.xe = graph.xMax;
      this.ys = this.ye = this.c/this.b;
    }
    else if (this.b === 0.0) {  // vertical
      this.xs = this.xe = this.c/this.a;
      this.ys = graph.yMin;
      this.ye = graph.yMax;
    }
    else {
      this.xs = graph.xMin;
      this.xe = graph.xMax;
      this.ys = (this.c - graph.xMin*this.a) / this.b;
      this.ye = (this.c - graph.xMax*this.a) / this.b;
    }
    this.line.setXY(this.xs, this.ys, this.xe, this.ye);
  };

  LPLine.prototype.reinit = function() {
    this.a = 0.0;
    this.b = 0.0;
    this.c = 0.0;
    this.xs = this.ys = this.xe = this.ye = INIT;
    this.line.setXY(INIT, INIT, INIT, INIT);
  };

  function ObjectiveFunction() {
    // call super class
    LPLine.call(this, 'of');
  }
  // inheritance
  ObjectiveFunction.prototype = Object.create(LPLine.prototype);

  // the following functions are used to snap the objective function line to a
  // vertex or vertices of the polygon

  ObjectiveFunction.prototype.seta = function(point) {
    if (point.x !== 0.0) {
      this.a = (this.c - this.b*point.y) / point.x;
      aSlider.setValue(this.a);
    }
  };

  ObjectiveFunction.prototype.setb = function(point) {
    if (point.y !== 0.0) {
      this.b = (this.c - this.a*point.x) / point.y;
      bSlider.setValue(this.b);
    }
  };

  ObjectiveFunction.prototype.setc = function(point) {
    this.c = this.a*point.x + this.b*point.y;
    cSlider.setValue(this.c);
  };

  ObjectiveFunction.prototype.setFunction = function(point0, point1) {
    var p0 = {x: this.xs, y: this.ys};
    var p1 = {x: p0.x + this.b, y: p0.y - this.a};
    var slope = getSlope(p0, p1);
    var newSlope = getSlope(point0, point1);
    var ratio;

    if (slope === newSlope) {
      // do nothing
    }
    else if (newSlope === null) {  // new line vertical
      this.b = 0.0;
      bSlider.setValue(this.b);
    }
    else if (newSlope === 0.0) {  // new line horizontal
      this.a = 0.0;
      aSlider.setValue(this.a);
    }
    else if (slope === null) {  // vertical
      this.b = -this.a / newSlope;
      bSlider.setValue(this.b);
    }
    else if (slope === 0.0) {  // horizontal
      this.a = -this.b * newSlope;
      aSlider.setValue(this.a);
    }
    else {
      ratio = newSlope/slope;
      p1.y = p1.y*ratio;
      p0.y = p0.y*ratio;
      this.a = p0.y - p1.y;
      aSlider.setValue(this.a);
    }
    this.setc(point0);
  };

  function Constraint() {
    // call super class
    LPLine.call(this, 'c');

    this.direction = 0;
    this.sign = '= ';
    this.p0 = new Point(graph, 0.0, 0.0, 2.0, 'c-point');
    this.p1 = new Point(graph, 0.0, 0.0, 2.0, 'c-point');
    _.invoke([this.p0, this.p1], 'setVisibility', false);
  }
  // inheritance
  Constraint.prototype = Object.create(LPLine.prototype);

  Constraint.prototype.setConstraint = function() {
    this.a = this.p0.y - this.p1.y;
    this.b = this.p1.x - this.p0.x;
    this.c = this.p1.x * this.p0.y - this.p0.x * this.p1.y;
    this.setLine();
  };

  Constraint.prototype.distanceTo = function(x, y) {
    var pm, pb, ix, iy;

    if (this.b === 0.0) {  // vertical
      return Math.abs(this.p0.x - x);
    }
    else if (this.a === 0.0) {  // horizonal
      return Math.abs(this.p0.y - y);
    }
    else {
      pm = this.b / this.a;  // perpendicular line
      pb = y - pm * x;
      ix = (this.c / this.b - pb) /  // intersection point
           (pm + this.a / this.b);
      iy = pm * ix + pb;
      return Math.sqrt((ix - x) * (ix - x) + (iy - y) * (iy - y));
    }
  };

  Constraint.prototype.setDirection = function(x, y) {
    this.direction = this.a * x + this.b * y < this.c ? -1 : 1;
  };

  Constraint.prototype.setSign = function(x, y) {
    if (this.b === 0.0) {  // vertical
      this.sign = x < this.p0.x ? U.STR.le + ' ' : this.sign = U.STR.ge + ' ';
    }
    else if (this.a === 0.0) {  // horizontal
      this.sign = y < this.p0.y ? U.STR.le + ' ' : this.sign = U.STR.ge + ' ';
    }
    else {
      this.sign = this.a * x + this.b * y < this.c ?
        U.STR.le + ' ' : this.sign = U.STR.ge + ' ';
    }
  };

  Constraint.prototype.getString = function() {
    if (Math.round(this.b) === 0.0) {  // vertical
      return ['x ', '', '', this.sign, (this.p0.x).toFixed(0)];
    } else if (Math.round(this.a) === 0.0) {  // horizontal
      return ['', '', 'y ', this.sign, (this.p0.y).toFixed(0)];
    } else {
      return [
        (this.a).toFixed(0) + ' x ',
        this.b < 0.0 ? ' - ' : ' + ',
        Math.abs(this.b).toFixed(0) + ' y ',
        this.sign,
        (this.c).toFixed(0)
      ];
    }
  };

  // Optimal Region Polygon

  function PolyPoint(x, y, index, isVertex) {
    this.x = x;
    this.y = y;
    this.index = index;
    this.isVertex = isVertex;
  }

  function LPPolygon(vertices) {
    this.initialVs = vertices;
    this.Vs = vertices;
    this.polygon = new Polygon(graph, vertices, 'polygon');
  }

  LPPolygon.prototype.checkCondition = function (v, a, b, c)  {
    return roundTo(a * v.x + b * v.y, 1000) >= roundTo(c, 1000);
  };

  LPPolygon.prototype.checkIntersection = function(v, a, b, c)  {
    return roundTo(a * v.x + b * v.y, 1000) === roundTo(c, 1000);
  };

  LPPolygon.prototype.intersectSegment = function(v0, v1, a, b, c) {
    // RED_FLAG: Assumes there is an intersection
    var t = (c - a * v0.x - b * v0.y) /
            (a * (v1.x - v0.x) + b * (v1.y - v0.y));
    var vertex = {
      x: v0.x + t * (v1.x - v0.x),
      y: v0.y + t * (v1.y - v0.y)
    };
    return vertex;
  };

  LPPolygon.prototype.update = function(constraint) {
    //Jerry's algorithm: We traverse the polygon checking if each vertex
    //satisfies the condition. When one vertex of a side satisfies the
    //condition and the other doesn't the line ax + by = c intesects the
    //segment. The point of intersection is a vertex on the updated polygon
    //and the vertices that don't satisfy the condition are removed from the
    //polygon.
    var newVs = [];
    var transitions = 0;
    var v0, v1, vNew;  // vertices
    var c0, c1, cBeg;  // conditions
    var i;

    // ax + by >= c
    var sign = constraint.direction;
    var a = sign * constraint.a;
    var b = sign * constraint.b;
    var c = sign * constraint.c;

    if (this.Vs.length === 0) {
      return;
    }

    v0 = {x: this.Vs[0], y: this.Vs[1]};
    c0 = this.checkCondition(v0, a, b, c);
    cBeg = c0;

    if (c0) {
      newVs.push(v0.x, v0.y);
    }

    for (i = 2; i < this.Vs.length; i+=2) {

      v1 = {x: this.Vs[i], y: this.Vs[i + 1]};
      c1 = transitions < 2 ? this.checkCondition(v1, a, b, c) : c0;

      if (c0) {
        if (c1) {
          // both vertices satisfy condition
          newVs.push(v1.x, v1.y);
        }
        else {
          // line ax + by = c intersects segment v0-v1
          if (!this.checkIntersection(v0, a, b, c)) {
            // if the intersection is at v0, it's already been added.
            vNew = this.intersectSegment(v0, v1, a, b, c);
            newVs.push(vNew.x, vNew.y);
          }
          c0 = false;
          transitions++;
        }
      }
      else {
        // v0 does not satisfy condition
        if (c1) {
          // line ax + by = c intersects segment v0-v1
          if (!this.checkIntersection(v1, a,b,c)) {
            vNew = this.intersectSegment(v0, v1, a, b, c);
            newVs.push(vNew.x, vNew.y);
          }
          newVs.push(v1.x, v1.y);
          c0 = true;
          transitions++;
        }
        // else do nothing
      }
      v0 = v1;
    }

    if (!cBeg && newVs.length >= 4) {
      // close the polygon
      newVs.push(newVs[0], newVs[1]);
    }
    this.polygon.setPoints(newVs);
    this.Vs = newVs;
  };

  LPPolygon.prototype.intersectionPoints = function(a, b, c) {
    //Jerry's algorithm: We traverse the polygon checking if each segment
    //intersects the line ax + by = c. The logic is the same as in update:
    //checking if the condition ax+by <= c is satisfied at each vertex.
    var intersectionPoints = [];
    var intersections = 0;
    var v0, v1;  // vertices
    var c0, c1;  // conditions
    var i;  // index counter
    var v;  // temp vertex

    if (this.Vs.length === 0) {
      return [];
    }

    v0 = new PolyPoint(this.Vs[0], this.Vs[1], 0, true);
    c0 = this.checkCondition(v0, a, b, c);

    if (c0) {
      if (this.checkIntersection(v0, a,b,c)) {
        intersectionPoints.push(v0);
        intersections++;
      }
    }

    for (i = 2; i < this.Vs.length; i+=2) {
      v1 = new PolyPoint(this.Vs[i], this.Vs[i + 1], i, true);
      c1 = this.checkCondition(v1, a, b, c);

      if (c0) {
        if (c1) {
          if (this.checkIntersection(v1, a,b,c)) {
            intersectionPoints.push(v1);
            intersections++;
          }
        }
        else {
          if (!this.checkIntersection(v0, a,b,c)) {
            v = this.intersectSegment(v0, v1, a, b, c);
            intersectionPoints.push(new PolyPoint(v.x, v.y, -1, false));
            intersections++;
          }
          c0 = false;
        }
      }
      else {
        if (c1) {
          if (this.checkIntersection(v1, a,b,c)) {
            intersectionPoints.push(v1);
          }
          else {
            v = this.intersectSegment(v0, v1, a, b, c);
            intersectionPoints.push(new PolyPoint(v.x, v.y, -1, false));
          }
          c0 = true;
          intersections++;
        }
        //else do nothing
      }
      if (intersections >= 2) {
        break;
      }
      v0 = v1;
    }
    return intersectionPoints;
  };

  LPPolygon.prototype.threshVs = function(a, b, c) {
    var points = [];
    var threshold = 0.25;
    var diff;
    var i, v;

    // search for polygon vertices within threshold of ax + by = c
    for (i = 0; i < this.Vs.length-2; i+=2) {
      diff = Math.abs(c - (a*this.Vs[i] + b*this.Vs[i+1]));
      if (diff < threshold) {
        v = new PolyPoint(this.Vs[i], this.Vs[i+1], i, true);
        v.diff = diff;
        points.push(v);
      }
    }
    // sort the points from least difference to greatest
    return _.sortBy(points, 'diff');
  };

  LPPolygon.prototype.VsAdjacent = function(v0, v1) {
    return (
      Math.abs(v1.index - v0.index) === 2 ||
      Math.abs(v1.index - v0.index) === this.Vs.length - 4);
  };

  LPPolygon.prototype.reinit = function() {
    this.Vs = this.initialVs;
    this.polygon.setPoints(this.Vs);
  };

  // Initialize + state related

  // The following functions are used by JSInput and have to be global
  function getState() {
    return JSON.stringify(state);
  }

  function setState(stateStr) {
    state = JSON.parse(stateStr);
    updateStateFromOutside();
  }

  function getGrade() {
    // The following return value may or may not be used to grade server-side.
    // If getState and setState are used, then the Python grader also gets
    // access to the return value of getState and can choose it instead to grade
    return JSON.stringify(state);
  }

  window.linearProgramming = {
    getState: getState,
    setState: setState,
    getGrade: getGrade
  };

  function updateStateFromOutside() {
    // TODO: Add state bindings
  }

  function updateStateFromInside() {
    // TODO: Add state bindings
  }

  initializeTool();

  function initializeTool() {
    // The try catch block checks if certain features are present.
    // If not, we exit and alert the user.
    try {
      U.testForFeatures(false, false, false, true); //Test for SVG only
    }
    catch (err) {
      window.alert(err.toString() + ' The tool is disabled.');
    }
    U.enableHelpLink('help-text', 'help-link');
    initTool();
  }

  function initTool() {
    var toolContainer = document.getElementById('daimp-tool-container');
    var sampleText, textxStart;  // to calculate text position
    var sliderCallbackList;
    var components, i;

    tool = new Tool(toolContainer); //Default size of 818x575

    U.setLanguage('kre');

    //// COMPONENT INITIALIZATION

    // Graph

    graph = new Graph(75, 50, 350, 350);
    graph.setPlottingBounds(0, 100, 0, 100);
    graph.setXGrid(20.0);
    graph.setYGrid(20.0);
    graph.setXLongTicks(20.0);
    graph.setYLongTicks(20.0);
    graph.setXShortTicks(10.0);
    graph.setYShortTicks(10.0);
    graph.setAutomaticXLabels(20.0, 0, true);
    graph.setAutomaticYLabels(20.0, 0, true);
    graph.setXAxis(0.0);
    graph.setYAxis(0.0);
    graph.setXText('x');
    graph.setYText('y');
    graph.setCrosshairs(true, true);
    tool.add(graph);

    // Text

    constraintsLabel = new Text(graph.x + graph.width + 95, graph.y + 20);
    constraintsLabel.addText(U.getText('Constraints'), 'c-text');
    tool.add(constraintsLabel);

    objFunctionLabel = new Text(
      constraintsLabel.x, graph.y + graph.height - 50);
    objFunctionLabel.addText(U.getText('Objective Function'), 'of-text');
    tool.add(objFunctionLabel);

    textxStart = constraintsLabel.x;
    // constraints text placeholders with max num characters
    sampleText = _.map(['100 x ', ' + ', '100 y ', U.STR.le + ' ', '10000'],
      function(str) {
        var text = new Text(textxStart, constraintsLabel.y + 40);
        text.addText(str, 'normal-text');
        text.setVisibility(false);
        tool.add(text);
        textxStart += text.text.getComputedTextLength();
        return text;
    });

    constraintsText = [];
    _.times(MAX_CONSTRAINTS, function(i) {
      var textGroup = _.map(['', '', '', '', ''], function(str, j) {
        var text = new Text(sampleText[j].x, constraintsLabel.y + 40 + i*25);
        text.addText(str, 'normal-text');
        tool.add(text);
        return text;
      });
      constraintsText.push(textGroup);
    });

    textxStart = objFunctionLabel.x;
    // objective function text placeholders with max num characters
    objFunctionText = _.map(['1.00x ', '+ ', '1.00y ', '= ', '100.00'],
      function(str) {
        var text = new Text(textxStart, objFunctionLabel.y + 40);
        text.addText(str, 'of-text');
        tool.add(text);
        textxStart += text.text.getComputedTextLength();
        return text;
    });

    // Sliders

    aSlider = new HSlider(
        objFunctionLabel.x, objFunctionLabel.y + 100, 200, 20);
    aSlider.setPlottingBounds(-1.0, 1.0);
    aSlider.setLongTicks(0.5);
    aSlider.setShortTicks(0.25);
    aSlider.setAutomaticLabels(0.5, 1);
    aSlider.setValue(0.0, 2);
    aSlider.setText('a');
    tool.add(aSlider);

    bSlider = new HSlider(aSlider.x, aSlider.y + 40, 200, 20);
    bSlider.setPlottingBounds(-1.0, 1.0);
    bSlider.setLongTicks(0.5);
    bSlider.setShortTicks(0.25);
    bSlider.setAutomaticLabels(0.5, 1);
    bSlider.setValue(0.0, 2);
    bSlider.setText('b');
    tool.add(bSlider);

    cSlider = new HSlider(bSlider.x, bSlider.y + 40, 200, 20);
    cSlider.setPlottingBounds(-50.0, 50.0);
    cSlider.setLongTicks(25.0);
    cSlider.setShortTicks(5.0);
    cSlider.setAutomaticLabels(25.0, 0);
    cSlider.setValue(0.0, 2);
    cSlider.setVisibility(false);
    cSlider.setText('c');
    tool.add(cSlider);

    _.each([aSlider, bSlider, cSlider], function(slider) {
      slider.snapToStep = true;
      slider.snapStep = 0.01;
    });

    // Readout Text

    xReadoutText = new Text(graph.x, aSlider.y + 8);
    xReadoutText.addText('x = ', 'normal-text');
    tool.add(xReadoutText);

    yReadoutText = new Text(xReadoutText.x, xReadoutText.y + 25);
    yReadoutText.addText('y = ', 'normal-text');
    tool.add(yReadoutText);

    // Checkbox

    removeCb = new Checkbox(
        xReadoutText.x + 140, xReadoutText.y - 12, 12, 12);
    removeCb.addText(U.getText('Remove Constraint'), 'disable-checkbox-label');
    removeCb.setFrontSquareClass('c-checkbox-front-square');
    tool.add(removeCb);

    // Button

    resetButton = new Button(removeCb.x, bSlider.y - 8, 80, 20);
    resetButton.addText(U.getText('Reset '));
    tool.add(resetButton);

    //// COMPONENT CALLBACK ASSIGNMENT

    // Graph

    graph.addEventListener('mouseover', function(mpos) {
      graphMove(mpos);
      setConstraintSign(mpos);
    });
    graph.addEventListener('mousemove', function(mpos) {
      graphMove(mpos);
      setConstraintSign(mpos);
    });
    graph.addEventListener('mousedown', function(mpos) {
      graphMove(mpos);
      if (removeCb.isChecked) {
        removeConstraint(mpos);
        removeCb.setChecked(false);
        dragEnabled = false;
      } else {
        graphDown(mpos);
        dragEnabled = true;
      }
    });
    graph.addEventListener('mousedrag', function(mpos) {
      graphMove(mpos);
      if (dragEnabled && !removeCb.isChecked) {
        graphDrag(mpos);
      }
    });
    graph.addEventListener('mouseout', graphOut);
    graph.addEventListener('mouseup', function(mpos) {
      var inBound = !(mpos.x === graph.xMin || mpos.x === graph.xMax ||
                      mpos.y === graph.yMin || mpos.y === graph.yMax);
      if (inBound) {
        graphMove(mpos);
      } else {
        graphOut(mpos);
      }
    });

    // Sliders

    sliderCallbackList = [aSliderMouseDown, bSliderMouseDown, cSliderMouseDown];

    _.each([aSlider, bSlider, cSlider], function(slider, i) {
      slider.addEventListener('mousedown', function() {
        _.throttle(sliderCallbackList[i](), 50);
      });
      slider.addEventListener('mousedrag', function() {
        _.throttle(sliderCallbackList[i](), 50);
      });
      slider.addEventListener('mouseup', function() {
        _.throttle(sliderCallbackList[i](), 50);
      });
    });

    // Checkbox
    removeCb.addEventListener('mousedown', function() {
      if (removeCb.isChecked) {
        if (stage !== 0 || ci <= 2) {
          removeCb.setChecked(false);
        }
      }
    });

    // Button

    resetButton.addEventListener('mousedown', reset);

    //// OBJECT INITIALIZATION

    polygon = new LPPolygon([
      0.0, 0.0, BIG_NUM, 0.0, BIG_NUM, BIG_NUM, 0.0, BIG_NUM, 0.0, 0.0]);

    for (i = 0; i < MAX_CONSTRAINTS; i++) {
      constraints.push(new Constraint());
    }
    constraints[0].p0.setPos(0, INIT);
    constraints[0].p1.setPos(0, graph.yMax - INIT);
    constraints[1].p0.setPos(INIT, 0);
    constraints[1].p1.setPos(graph.xMax - INIT, 0);
    _.times(2, function(i) {
      constraints[i].setConstraint();
      constraints[i].setSign(1.0, 1.0);
      constraints[i].setDirection(1.0, 1.0);
      setConstraintText(i);
    });

    objFunction = new ObjectiveFunction();
    setObjFunctionText();

    ipLine = new Line(graph, INIT, INIT, INIT, INIT, 'ip-line');

    ipPoints = [
      new Diamond(graph, INIT, INIT, 3.0, 'ip-point'),
      new Diamond(graph, INIT, INIT, 3.0, 'ip-point')
    ];
    tool.finalize();
  }  // end initTool()

  function reset() {
    var i;

    removeCb.setChecked(false);
    removeCb.setClass(0, 'disable-checkbox-label');

    for (i = 2; i < ci; i++) {
      constraints[i].reinit();
      resetConstraintText(i);
    }
    if (stage > 0) { // check for constraints in progress
      constraints[i].reinit();
      resetConstraintText(i);
      _.invoke([constraints[i].p0, constraints[i].p1], 'setVisibility', false);
    }
    polygon.reinit();
    objFunction.reinit();
    setObjFunctionText();
    setIntersectionLine();

    aSlider.setValue(0.0);
    bSlider.setValue(0.0);
    cSlider.setValue(0.0);
    cSlider.setVisibility(false);

    stage = 0;
    ci = 2;
  }

  function removeConstraint(mpos) {
    var i = closestConstraint(mpos.x, mpos.y);
    var c = constraints[i];
    ci--;
    // move each constraint > i up the list
    for (; i < ci; i++) {
      constraints[i] = constraints[i+1];
      setConstraintText(i);
    }
    c.reinit();
    constraints[ci] = c;
    resetConstraintText(ci);

    // recalculate polygon points
    polygon.reinit();
    for (i = 2; i < ci; i++) {
      polygon.update(constraints[i]);
    }
    setObjFunctionCoeffToVs(C); //check obj function with new polygon Vs
    setObjFunction();
    if (ci <= 2) {
      removeCb.setClass(0, 'disable-checkbox-label');
    }
  }

  function closestConstraint(x, y) {
    var minDist = Number.MAX_VALUE;
    var dist;
    var result;
    var i;
    for (i = 2; i < ci; i++) {
      dist = constraints[i].distanceTo(x, y);
      if (dist <= minDist) {
        minDist = dist;
        result = i;
      }
    }
    return result;
  }

  function graphMove(mpos) {
    xReadoutText.setText(0, 'x = ' + (mpos.x).toFixed(0));
    yReadoutText.setText(0, 'y = ' + (mpos.y).toFixed(0));
  }

  function graphOut(mpos) {
    xReadoutText.setText(0, 'x =');
    yReadoutText.setText(0, 'y =');
  }

  function graphDown(mpos) {
    switch(stage) {
      case 0:
        if (ci < MAX_CONSTRAINTS) {
          stage++;
          constraints[ci].p0.setVisibility(true);
          graphDrag(mpos);
          removeCb.setClass(0, 'disable-checkbox-label');
        }
        break;
      case 1:
        stage++;
        constraints[ci].p1.setVisibility(true);
        graphDrag(mpos);
        break;
      case 2:
        stage = 0;
        _.invoke(
          [constraints[ci].p0, constraints[ci].p1], 'setVisibility', false);
        constraints[ci].setDirection(mpos.x, mpos.y);
        polygon.update(constraints[ci]);
        setObjFunctionCoeffToVs(C);
        setObjFunction();
        ci++;
        removeCb.setClass(0, 'c-checkbox-label');
        break;
    }
  }

  function graphDrag(mpos) {
    if (stage === 1) {
      constraints[ci].p0.setPos(Math.round(mpos.x), Math.round(mpos.y));
    }
    if (stage === 2) {
      constraints[ci].p1.setPos(Math.round(mpos.x), Math.round(mpos.y));
      constraints[ci].setConstraint();
      setConstraintText(ci);
    }
  }

  function aSliderMouseDown() {
    objFunction.a = aSlider.value;
    if (objFunction.a === 0.0 && objFunction.b === 0.0) {
      objFunction.c = 0.0;
      cSlider.setVisibility(false);
      cSlider.setValue(0.0);
    } else {
      cSlider.setVisibility(true);
      setObjFunctionCoeffToVs(A);
    }
    setObjFunction();
  }

  function bSliderMouseDown() {
    var points;
    objFunction.b = bSlider.value;
    if (objFunction.a === 0.0 && objFunction.b === 0.0) {
      objFunction.c = 0.0;
      cSlider.setVisibility(false);
      cSlider.setValue(0.0);
    } else {
      cSlider.setVisibility(true);
      setObjFunctionCoeffToVs(B);
    }
    setObjFunction();
  }

  function cSliderMouseDown() {
    objFunction.c = cSlider.value;
    setObjFunctionCoeffToVs(C);
    setObjFunction();
  }

  function setObjFunction() {
    objFunction.setLine();
    setObjFunctionText();
    setIntersectionLine();
  }

  function setObjFunctionCoeffToVs(coeff) {
    var threshVs = polygon.threshVs(
      objFunction.a, objFunction.b, objFunction.c
    );
    var p0, p1;

    if (threshVs.length >= 2 && polygon.VsAdjacent(threshVs[0], threshVs[1])) {
      p0 = threshVs[0];
      p1 = threshVs[1];
      if (p0.diff !== 0.0 || p1.diff !== 0.0) {
        objFunction.setFunction(p0, p1);  // set the line
      }
    } else if (threshVs.length >= 1) {
      p0 = threshVs[0];
      if (p0.diff !== 0.0) {  // point is not intersected
        if (coeff === A) {
          objFunction.seta(p0);
        }
        else if (coeff === B) {
          objFunction.setb(p0);
        }
        else {
          objFunction.setc(p0);
        }
      }
    }
  }

  function setIntersectionLine() {
    var intersectionPoints = [];
    var p0, p1;

    if (objFunction.a === 0.0 && objFunction.b === 0.0) {
      if (polygonIntersected) {
        polygonIntersected = false;
        polygon.polygon.setClass('polygon');
        ipVisibility(false);
      }
      return;
    }

    intersectionPoints = polygon.intersectionPoints(
      objFunction.a, objFunction.b, objFunction.c
    );
    if (intersectionPoints.length >= 1) {
      p0 = intersectionPoints[0];
      if (!polygonIntersected) {
        polygonIntersected = true;
        polygon.polygon.setClass('polygon-lit');
        ipVisibility(true);
      }
      if (intersectionPoints.length >= 2) {
        p1 = intersectionPoints[1];
        if ((p0.isVertex && p1.isVertex) && polygon.VsAdjacent(p0, p1)) {
          // intersection occurs along line segment
          ipLine.setClass('ip-line-lit');
        }
        else {
          ipLine.setClass('ip-line');
        }
        // set p1 vertex and intersection line
        ipPoints[1].setXY(p1.x, p1.y);
        ipLine.setXY(p0.x, p0.y, p1.x, p1.y);
      } else {
        // set outside of graphing bounds
        ipPoints[1].setXY(INIT, INIT);
        ipLine.setXY(INIT, INIT, INIT, INIT);
      }
      // set p0 vertex
      ipPoints[0].setXY(p0.x, p0.y);
    }
    else if (polygonIntersected) {
      // polygon is no longer intersected, set to false
      polygonIntersected = false;
      polygon.polygon.setClass('polygon');
      ipVisibility(false);
    }
  }

  function ipVisibility(bool) {
    ipLine.setVisibility(bool);
    _.invoke(ipPoints, 'setVisibility', bool);
  }

  // update text

  function setObjFunctionText() {
    objFunctionText[A].setText(0, objFunction.a.toFixed(2) + 'x ');
    objFunctionText[OPER].setText(0, objFunction.b < 0.0 ? ' - ' : '+ ');
    objFunctionText[B].setText(0, Math.abs(objFunction.b).toFixed(2) + 'y ');
    objFunctionText[C].setText(0, (objFunction.c).toFixed(2));

    objFunctionText[A].setPosition(
      objFunctionText[OPER].x - objFunctionText[A].text.getComputedTextLength(),
      objFunctionText[A].y);
  }

  function setConstraintText(i) {
    var str = constraints[i].getString();
    var txt = constraintsText[i];
    _.each(txt, function(t, j) {
      t.setText(0, str[j]);
    });
    txt[A].setPosition(
      txt[OPER].x - txt[A].text.getComputedTextLength(), txt[A].y);
    txt[B].setPosition(
      txt[SIGN].x - txt[B].text.getComputedTextLength(), txt[B].y);
  }

  function setConstraintSign(mpos) {
    if (stage === 2) {
      constraints[ci].setSign(mpos.x, mpos.y);
      constraintsText[ci][SIGN].setText(0, constraints[ci].sign);
    }
  }

  function resetConstraintText(i) {
    _.each(constraintsText[i], function(text) {
      constraints[i].sign = '= ';
      text.setText(0, '');
    });
  }

  // calculations

  function getSlope(point0, point1) {
    var xChange = (point1.x-point0.x);
    var yChange = (point1.y-point0.y);
    if (yChange === 0.0) {
      return 0.0; // horizontal
    }
    else if (xChange === 0.0) {
      return null; // vertical, slope undefined
    }
    else {
      return yChange / xChange;
    }
  }

  function roundTo(n, dec) {
    return Math.round(n*dec)/dec;
  }
});
